/*
 *  Autor:    Edward Martin
 *
 *  Objetivo: Gerar os eventos de forma mais rpida possvel:
 *            1) Nova conexo
 *            2) Chegada de dados de uma conexo (um "pipe")
 *            3) Aps decorrido um tempo, que pode ser de 1 a 300 segundos
 *
 *  Processar conexo - nvel de objetos (ver detalhes em eventos.h):
 *            Os objetos tem nveis, de 0 a 5.
 *            Quando so criados, so nveis 4, mas isso pode ser mudado.
 *            Objetos nveis 5: Mensagens s sero enviadas quando o buffer
 *                      de sada ficar cheio (til para telnet)
 *            Objetos nveis 4: Mensagens so enviadas, porm se o buffer
 *                      de sada ficar cheio, a funo informa que
 *                      no conseguiu enviar.
 *            Objetos nveis 3-0: Mensagens so enviadas e a funo no
 *                      retorna at a mensagem caber no buffer de sada.
 *                      Enquanto isso, eventos de objetos de menor nvel
 *                      podero ser chamados.
 *
 *  Processar conexo - voltar mensagens
 *            Permite repetir as mensagens enviadas nos ltimos 3 segundos.
 *            til para quem conecta usando telnet falado, para evitar
 *            perder mensagens.
 *
 *  Modo de usar:
 *            Criar classes derivadas das classes que geram eventos.
 *
 *  Obs.:     As classes prevem que qualquer objeto pode ser apagado
 *            em qualquer evento.
 */

#include <stdio.h>
#include <stdlib.h>

#ifdef __WIN32__
 #include <windows.h>
 #include <winsock.h>
 #define close closesocket
 #define acceptarg3 int
#else
 #include <unistd.h>     // read e write
 #include <fcntl.h>
 #include <string.h>
 #include <sys/time.h>
 #include <sys/types.h>
 #include <sys/socket.h> // Comunicao
 #include <signal.h>
 #include <netinet/in.h> // Contm a estrutura sockaddr_in
 #include <sys/un.h>     // Contm a estrutura sockaddr_un
 #include <errno.h>
 #include "def.h"
#endif
#include <time.h>
#include "eventos.h"
#include "misc.h"
//#define acceptarg3 socklen_t // Alteranativa para "def.h"

/*
Para testar processamento dos eventos do relgio

class Tteste : public TRelogio
{
public:
    Tteste();
    void rEvento();
    bool evento;
};
Tteste::Tteste() { evento=false; }
void Tteste::rEvento() { evento=true; rTempo(4); }

    Tteste t[12];
    for (int a=0; a<12; a++)
        TRelogio::ProcEventos(1);
    for (int a=0; a<12; a++)
    {
        t[a].rTempo(a);
        printf("[%d %d]", (int)t[a].Indice, (int)t[a].Indice2);
    }
    putchar('\n');
    for (int b=0; b<14; b++)
    {
        for (int a=0; a<12; a++)
            t[a].evento=false;
        TRelogio::ProcEventos(1);
        //printf("%2d",b);
        printf("%2d %2d", TRelogio::TempoAtual1, TRelogio::TempoAtual2);
        for (int a=0; a<12; a++)
        {
            printf("  %c", t[a].evento ? 'E' : ' ');
            putchar('0'+t[a].rTempo());
            //printf("[%d %d]", (int)t[a].Indice, (int)t[a].Indice2);
        }
        putchar('\n');
    }
*/


//------------------------------------------------------------------------------
// Deve ser chamado para terminar o programa
void Terminar(int retorno)
{
#ifdef __WIN32__
    WSACleanup();
#endif
    exit(retorno);
}

//------------------------------------------------------------------------------
TRelogio * TRelogio::Inicio[TRELOGIO_MAX*3];
int TRelogio::TempoAtual1=0;
int TRelogio::TempoAtual2=TRELOGIO_MAX*2;
TComunica * TComunica::Inicio[6];
TComunica * TComunica::Fim[6];
TComunica * TComunica::PontAtual[6];
TComunica * TComunica::PontProximo[6];
char TComunica::NivelProc=6;    // Deve ser inicializado com o nmero de nveis
                                // de TComunica (6)
TServidor * TServidor::Inicio=0;
TServidor * TServidor::PontAtual;
TServidor * TServidor::PontProximo;
static bool Processando=false;  // Verdadeiro se ProcessarIni() estiver sendo executado
static bool Term=false;         // Verdadeiro se ProcessarIni() deve terminar
static bool Inic=true;          // Verdadeiro se classes no foram preparadas
static struct timeval tatual;   // Tempo atual
long  Tempo;                    // Tempo atual (segundos), usado nos eventos
unsigned long TempoIni=0;       // Para contar tempo decorrido (varinc e vardec)
short Dia,Hora,Min;             // Dia,Hora e minuto atuais
unsigned long HoraReg;          // Usado no registro

//------------------------------------------------------------------------------
TRelogio::TRelogio()
{
    if (::Inic)
        ProcessarPrepara();
    Anterior=Proximo=0;
    Indice=-1;
}

//------------------------------------------------------------------------------
TRelogio::~TRelogio()
{
    if (Indice<0)
        return;
    (Anterior ? Anterior->Proximo : Inicio[Indice])=Proximo;
    if (Proximo)
        Proximo->Anterior=Anterior;
}

//------------------------------------------------------------------------------
void TRelogio::rEvento()
{
}

//------------------------------------------------------------------------------
bool TRelogio::rPendEvento()
{
    return (Indice>=0);
}

//------------------------------------------------------------------------------
int TRelogio::rTempo(void)
{
    if (Indice<0)
        return 0;
    if (Indice<TRELOGIO_MAX*2)
        return (TRELOGIO_MAX*2+Indice-TempoAtual1)%(TRELOGIO_MAX*2);
    return ((TRELOGIO_MAX+Indice-TempoAtual2)%TRELOGIO_MAX)*TRELOGIO_MAX
           + Indice2 + (TRELOGIO_MAX*2-TempoAtual1-1)%TRELOGIO_MAX+1;
}

//------------------------------------------------------------------------------
int TRelogio::rTempo(int NovoTempo)
{
    if (NovoTempo>TRELOGIO_MAX*TRELOGIO_MAX)
        NovoTempo=TRELOGIO_MAX*TRELOGIO_MAX;
    if (Indice>=0)              // Retira da lista de tempos
    {
        (Anterior ? Anterior->Proximo : Inicio[Indice])=Proximo;
        if (Proximo)
            Proximo->Anterior=Anterior;
    }
    if (NovoTempo<=0)        // No gerar evento
    {
        Indice=-1;
        Anterior=Proximo=0;
        return 0;
    }
    if (NovoTempo<TRELOGIO_MAX*2) // Acerta variveis
    {
        Indice=(TempoAtual1+NovoTempo)%(TRELOGIO_MAX*2);
        Indice2=0;
    }
    else
    {
        Indice2 = NovoTempo - (TRELOGIO_MAX*2-TempoAtual1-1)%TRELOGIO_MAX - 1;
        Indice  = (TRELOGIO_MAX + TempoAtual2 + Indice2/TRELOGIO_MAX)
                        % TRELOGIO_MAX + TRELOGIO_MAX*2;
        Indice2 = Indice2%TRELOGIO_MAX;
    }
    Anterior=0;        // Coloca na lista de tempos
    Proximo=Inicio[Indice];
    Inicio[Indice]=this;
    if (Proximo)
        Proximo->Anterior=this;
    return NovoTempo;
}

//------------------------------------------------------------------------------
void TRelogio::ProcEventos(int TempoDecorrido)
{
    TRelogio * rpont=0;
    int ind;
    while (TempoDecorrido>0)
    {
        TempoAtual1=(TempoAtual1+1)%(TRELOGIO_MAX*2);
        TempoDecorrido--;
        if (TempoAtual1==0 || TempoAtual1==TRELOGIO_MAX)
        {
            TempoAtual2++;
            if (TempoAtual2>=TRELOGIO_MAX*3)
                TempoAtual2=TRELOGIO_MAX*2;
            while ( (rpont=Inicio[TempoAtual2]) != 0 )
            {
                Inicio[TempoAtual2]=rpont->Proximo;
                if (rpont->Proximo)
                    rpont->Proximo->Anterior=0;
                rpont->Proximo=0;
                ind = (rpont->Indice2 + TempoAtual1 + TRELOGIO_MAX)
                                % (TRELOGIO_MAX*2);
                rpont->Indice = ind;
                rpont->Proximo = Inicio[ind];
                rpont->Anterior = 0;
                if (Inicio[ind])
                    Inicio[ind]->Anterior = rpont;
                Inicio[ind] = rpont;
            }
        }
        while ( (rpont=Inicio[TempoAtual1]) != 0 )
        {
            Inicio[TempoAtual1]=rpont->Proximo;
            if (rpont->Proximo)
                rpont->Proximo->Anterior=0;
            rpont->Proximo=0;
            rpont->Indice=-1;
            rpont->rEvento();
        }
    }
}

//------------------------------------------------------------------------------
TComunica::TComunica()
{
    if (::Inic)
        ProcessarPrepara();
    cSockEnv=cSockRec=-1;
    Leitura=Escrita=0;
    Anterior=0;
    Proximo=Inicio[4];
    Inicio[4]=this;
    (Proximo ? Proximo->Anterior : Fim[4])=this;
    Fechou=false;
    NivelEnvia=NivelEvento=4;
    Esc0Seg=Esc1Seg=Esc2Seg=Esc3Seg=-1;
    EscTempo=tatual.tv_sec;
}

//------------------------------------------------------------------------------
TComunica::~TComunica()
{
    if (PontAtual[NivelEvento]==this)
        PontAtual[NivelEvento]=0;
    if (PontProximo[NivelEvento]==this)
        PontProximo[NivelEvento]=Proximo;
    cFechaSock();
    (Anterior ? Anterior->Proximo : Inicio[NivelEvento])=Proximo;
    (Proximo ? Proximo->Anterior : Fim[NivelEvento])=Anterior;
}

//------------------------------------------------------------------------------
void TComunica::cEvento(const char * buffer, int tamanho)
{
}

//------------------------------------------------------------------------------
void TComunica::cFechaSock(bool enviar)
{
    if (cSockEnv>=0)
    {
        if (enviar && NivelEnvia!=5 && Leitura!=Escrita)
            cEnviaPend();
        close(cSockEnv);
    }
    if (cSockRec>=0 && cSockRec!=cSockEnv)
        close(cSockRec);
    cSockEnv=cSockRec=-1;
    Leitura=Escrita=0;
    Esc0Seg=Esc1Seg=Esc2Seg=Esc3Seg=-1;
    EscTempo=tatual.tv_sec;
    Fechou=false;
}

//------------------------------------------------------------------------------
char TComunica::cNivelEnvia(void)
{
    return NivelEnvia;
}

//------------------------------------------------------------------------------
char TComunica::cNivelEvento(void)
{
    return NivelEvento;
}

//------------------------------------------------------------------------------
void TComunica::cMudaNivel(char envia, char evento)
{
    if ((unsigned char)envia<6)
        NivelEnvia=envia;
    if (evento==NivelEvento || (unsigned char)evento>=6)
        return;
    if (PontProximo[NivelEvento]==this)  // Retira da lista de nveis
        PontProximo[NivelEvento]=Proximo;
    (Anterior ? Anterior->Proximo : Inicio[NivelEvento])=Proximo;
    (Proximo ? Proximo->Anterior : Fim[NivelEvento])=Anterior;
    NivelEvento=evento;         // Acerta o nvel
    Proximo=0;                  // Coloca na lista
    Anterior=Fim[NivelEvento];
    Fim[NivelEvento]=this;
    (Anterior ? Anterior->Proximo : Inicio[NivelEvento])=this;
}

//------------------------------------------------------------------------------
bool TComunica::cEnviaPend(void)
{
    int resposta,escrever;
    if (cSockEnv<0 || Leitura==Escrita)
        return (Leitura==Escrita);
    while (true)
    {
        escrever=Escrita-Leitura;
        if (escrever<0)
            escrever=COMUNICA_BUF-Leitura;
#ifdef __WIN32__
        resposta=send(cSockEnv, (const char *)&Buffer[Leitura], escrever, 0);
        if (resposta==0)
            resposta=-1;
        else if (resposta==SOCKET_ERROR)
        {
            resposta=WSAGetLastError();
            resposta = (resposta==WSAEWOULDBLOCK ? 0 : -1);
        }
#else
        resposta=send(cSockEnv, &Buffer[Leitura], escrever, 0);
        if (resposta<=0)
        {
            if (resposta<0 && (errno==EINTR || errno==EWOULDBLOCK || errno==ENOBUFS))
                resposta=0;
            else
                resposta=-1;
        }
#endif
    //printf("ENV1");
    //for (int a=0; a<escrever; a++)
    //  printf(" %d", (unsigned char)Buffer[Leitura+a]);
    //putchar('\n');
    //fflush(stdout);
        if (resposta>0)
        {
// if (NivelEnvia>=4)   // Mostra o que est enviando
// {
//  printf("Enviando [");
//  for (int a=Leitura; a<Leitura+resposta; a++)
//   if ((unsigned char)Buffer[a]>=' ')
//    putchar(Buffer[a]);
//   else
//    printf("] %d [", Buffer[a]);
//  puts("]");
// }
            Leitura=(Leitura+resposta)%COMUNICA_BUF;
            if (Leitura==Escrita)
                return true;
        }
        if (resposta<0) // Conexo terminada
        {
            if (cSockEnv!=cSockRec)
                close(cSockEnv);
            cSockEnv=-1;
            Fechou=true;
            return false;
        }
        if (resposta!=escrever)
            return false;
    }
}

//------------------------------------------------------------------------------
bool TComunica::cEnvia(const void * buffer, int tamanho)
{
    int rasc;
// Verifica se tamanho vlido
    if (tamanho<0)
        tamanho=strlen((const char *)buffer);
    if (tamanho==0)
        return true;
    if (tamanho>=COMUNICA_BUF)
        return false;
   // printf("ENV");
   // for (int a=0; a<tamanho; a++)
   //   printf(" %d", ((unsigned char*)buffer)[a]);
   // putchar('\n');
   // fflush(stdout);
//if (NivelEnvia!=NivelEvento)
// printf("[TComunica]  NivelEnvia=%d  NivelEvento=%d  NivelProc=%d\n",
//        NivelEnvia, NivelEvento, NivelProc);
    if (NivelEnvia<4 && NivelEnvia>NivelProc)
    {
        puts("[TComunica] Erro: cEnvia() foi chamado de cEvento() de menor nvel");
        fflush(stdout);
#ifdef FUNC_MEM1
        char * p=0;
        *p=0;
#endif
        return false;
    }
// Verifica se tem espao no buffer
    if (tamanho >= (COMUNICA_BUF+Leitura-Escrita-1)%COMUNICA_BUF)
    {
        if (cSockEnv<0)
            return false;
// Nvel >= 4: No espera at buffer ficar vazio
        if (NivelEnvia>=4)
        {
            NivelEnvia=4;
            cEnviaPend();
            if (tamanho >= (COMUNICA_BUF+Leitura-Escrita-1)%COMUNICA_BUF)
                return false;
        }
// Nvel < 4: Espera buffer ficar vazio, pode chamar outros eventos
        else
        {
            fd_set set_entrada,set_saida;
            TComunica * cpont;
            char cnivel=NivelEnvia;
            PontAtual[(int)cnivel]=this;
            while (!cEnviaPend())
            {
        // Verifica se conexo terminou
                if (cSockEnv<0)
                    return false;
        // Prepara para select()
                FD_ZERO(&set_entrada);
                FD_ZERO(&set_saida);
                FD_SET(cSockEnv, &set_saida);
                for (rasc=0; rasc<cnivel; rasc++)
                    for (cpont=TComunica::Inicio[rasc]; cpont; cpont=cpont->Proximo)
                        if (cpont->cSockRec>=0)
                            FD_SET(cpont->cSockRec, &set_entrada);
                select(FD_SETSIZE, &set_entrada, &set_saida, 0, 0);
        // Verifica se j pode escrever na sada
                if (FD_ISSET(cSockEnv, &set_saida))
                    continue;
        // Chama outros eventos
                ProcEventos(&set_entrada, cnivel);
                if (PontAtual[(int)cnivel]==0)
                    return false;
            }
        }
    }
// Acerta ponteiros para mensagens anteriores
    switch ((long)tatual.tv_sec-EscTempo)
    {
    case 0:
    //    if (Esc0Seg<0)
    //        Esc0Seg=Escrita;
        break;
    case 1:
        Esc3Seg=Esc2Seg;
        Esc2Seg=Esc1Seg;
        Esc1Seg=Esc0Seg;
        Esc0Seg=Escrita;
        break;
    case 2:
        Esc3Seg=Esc1Seg;
        Esc2Seg=Esc0Seg;
        Esc1Seg=-1;
        Esc0Seg=Escrita;
        break;
    case 3:
        Esc3Seg=Esc0Seg;
        Esc2Seg=Esc1Seg=-1;
        Esc0Seg=Escrita;
        break;
    default:
        Esc1Seg=Esc2Seg=Esc3Seg=-1;
        Esc0Seg=Escrita;
    }
    EscTempo=tatual.tv_sec;
    while (true)
    {
        if (Esc3Seg>=0)
            if (tamanho >= (COMUNICA_BUF+Esc3Seg-Escrita-1)%COMUNICA_BUF)
                Esc3Seg=-1;
            else break;
        if (Esc2Seg>=0)
            if (tamanho >= (COMUNICA_BUF+Esc2Seg-Escrita-1)%COMUNICA_BUF)
                Esc2Seg=-1;
            else break;
        if (Esc1Seg>=0)
            if (tamanho >= (COMUNICA_BUF+Esc1Seg-Escrita-1)%COMUNICA_BUF)
                Esc1Seg=-1;
            else break;
        if (Esc0Seg>=0)
            if (tamanho >= (COMUNICA_BUF+Esc0Seg-Escrita-1)%COMUNICA_BUF)
                Esc0Seg=-1;
        break;
    }
// Coloca dados no buffer
    rasc=(tamanho<COMUNICA_BUF-Escrita ? tamanho : COMUNICA_BUF-Escrita);
    memcpy(&Buffer[Escrita], buffer, rasc);
    Escrita=(Escrita+rasc)%COMUNICA_BUF;
    tamanho-=rasc;
    if (tamanho)
    {
        memcpy(Buffer, ((char*)buffer)+rasc, tamanho);
        Escrita=tamanho;
    }
    return true;
}

//------------------------------------------------------------------------------
void TComunica::cVoltaMens(void)
{
    short pont=-1;
    switch ((long)tatual.tv_sec-EscTempo)
    {
    case 0:
        if (Esc3Seg>=0)
        {
            pont=Esc3Seg;
            break;
        }
    case 1:
        if (Esc2Seg>=0)
        {
            pont=Esc2Seg;
            break;
        }
    case 2:
        if (Esc1Seg>=0)
        {
            pont=Esc1Seg;
            break;
        }
    case 3:
        if (Esc0Seg>=0)
            pont=Esc0Seg;
    }
    if (pont>=0)
        Leitura=pont;
    Esc0Seg=Esc1Seg=Esc2Seg=Esc3Seg=-1;
    EscTempo=tatual.tv_sec;
}

//------------------------------------------------------------------------------
void TComunica::cLimpaMens(void)
{
    Esc0Seg=Esc1Seg=Esc2Seg=Esc3Seg=-1;
    EscTempo=tatual.tv_sec;
}

//------------------------------------------------------------------------------
char * TComunica::cBuffer(void)
{
    return Buffer;
}

//------------------------------------------------------------------------------
void TComunica::ProcEventos(fd_set * set_entrada, char TotalNiveis)
{
    char buffer[COMUNICA_REC]; // Buffer para receber dados
    int resposta;              // Resposta ao receber dados
    int cont;
    char nivel=NivelProc;
    TComunica * cpont;
    for (cont=0; cont<TotalNiveis; cont++)
    {
        NivelProc=cont;
        for (cpont=TComunica::Inicio[cont]; cpont;)
        {
            PontProximo[cont]=cpont->Proximo;
            PontAtual[cont]=cpont;
            while (cpont->cSockRec>=0 && FD_ISSET(cpont->cSockRec, set_entrada))
            {
#ifdef __WIN32__
                resposta=recv(cpont->cSockRec, (char *)buffer, sizeof(buffer), 0);
                if (resposta==0)
                    resposta=-1;
                else if (resposta==SOCKET_ERROR)
                {
                    resposta=WSAGetLastError();
                    resposta = (resposta==WSAEWOULDBLOCK ? 0 : -1);
                }
#else
                resposta=recv(cpont->cSockRec, buffer, sizeof(buffer), 0);
                if (resposta<=0)
                {
                    if (resposta<0 && (errno==EINTR || errno==EWOULDBLOCK || errno==ENOBUFS))
                        resposta=0;
                    else
                        resposta=-1;
                }
#endif
                if (resposta<0)
                {
                    cpont->cFechaSock(false);
                    cpont->cEvento(0,0);
                }
                else if (resposta>0)
                    cpont->cEvento(buffer, resposta);
                if (TComunica::PontAtual[cont]==0 || resposta!=sizeof(buffer))
                    break;
            }
            if (TComunica::PontAtual[cont])
                if (cpont->Fechou)
                {
                    cpont->cFechaSock(false);
                    cpont->cEvento(0,0);
                }
            cpont=PontProximo[cont];
        }
    }
    NivelProc=nivel;
}

//------------------------------------------------------------------------------
TServidor::TServidor()
{
    if (::Inic)
        ProcessarPrepara();
    Proximo=Inicio;
    Anterior=0;
    Inicio=this;
    if (Proximo)
        Proximo->Anterior=this;
    Sock=-1;
    sProcessar=true;
}

//------------------------------------------------------------------------------
TServidor::~TServidor()
{
    if (Sock>=0)
        close(Sock);
    if (PontAtual==this)
        PontAtual=0;
    if (PontProximo==this)
        PontProximo=Proximo;
    (Anterior ? Anterior->Proximo : Inicio)=Proximo;
    if (Proximo)
        Proximo->Anterior=Anterior;
}

//------------------------------------------------------------------------------
char * TServidor::sAtivar(unsigned short * Porta)
{
    static char Retorno[100];
    const char *pont;
    struct sockaddr_in strSock;
    while (true)
    {
#ifdef __WIN__
        BOOL   iopcoes;
#else
        int    iopcoes;
#endif
        acceptarg3 stamanho;
        if (Sock>=0)
            close(Sock);
        pont="socket()";
#ifdef __WIN__
        if ( (Sock = socket(PF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET )
#else
        if ( (Sock = socket(PF_INET, SOCK_STREAM, 0)) <0 )
#endif
            break;
        iopcoes=1;
        stamanho=sizeof(iopcoes);
#ifdef __WIN32__
        //setsockopt(Sock, SOL_SOCKET, SO_REUSEADDR, (const char*)(void*)&iopcoes, stamanho);
        memset(&(strSock.sin_zero), 0, 8);
        strSock.sin_family = AF_INET;
        strSock.sin_port = htons((u_short)*Porta);
#else
        setsockopt(Sock, SOL_SOCKET, SO_REUSEADDR, &iopcoes, stamanho);
        memset(&(strSock.sin_zero), 0, 8);
        strSock.sin_family = AF_INET;
        strSock.sin_port = htons(*Porta);
#endif
        strSock.sin_addr.s_addr = htonl(INADDR_ANY);
        if (*Porta)
        {
            pont="bind()";
#ifdef __WIN32__
            if ( bind(Sock, (LPSOCKADDR)&strSock, sizeof(strSock)) )
#else
            if ( bind(Sock, (struct sockaddr *)&strSock,
                                sizeof(struct sockaddr))<0)
#endif
                break;
        }
        pont="listen()";
#ifdef __WIN32__
        if (listen(Sock, 15))
#else
        if (listen(Sock, 15)<0)
#endif
            break;
        if (*Porta==0)
        {
            acceptarg3 tam=sizeof(struct sockaddr);
            if (getsockname(Sock,(struct sockaddr *)&strSock,&tam)==0)
                *Porta=ntohs(strSock.sin_port);
        }
#ifdef __WIN32__
        unsigned long argp=1; // 0=bloquear  1=nao bloquear
        if ( ioctlsocket(Sock, FIONBIO, &argp) !=0 )
            break;
#else
        fcntl(Sock, F_SETFL, O_NONBLOCK);
#endif
        return 0;
    }
    mprintf(Retorno,100,"%s: %s",pont,strerror(errno));
    if (Sock>=0)
        close(Sock);
    Sock=-1;
    return Retorno;
}

//------------------------------------------------------------------------------
void TServidor::sDesativar(void)
{
    if (Sock>=0)
        close(Sock);
    Sock=-1;
}

//------------------------------------------------------------------------------
void TServidor::sProcTodos(bool Processar)
{
    for (TServidor * spont=Inicio; spont; spont=spont->Proximo)
        spont->sProcessar=Processar;
}

//------------------------------------------------------------------------------
void TServidor::sEvento(short EvSock, struct sockaddr_in * localSockStr)
{
    close(EvSock);
}

//------------------------------------------------------------------------------
void TServidor::ProcEventos(fd_set * set_entrada)
{
    acceptarg3 tamsock;
    int localSocket;
    struct sockaddr_in SockStr_in;
    TServidor * spont=Inicio;
    while (spont)
    {
        PontProximo=spont->Proximo;
        if (FD_ISSET(spont->Sock, set_entrada))
        {
            PontAtual=spont;
            while (spont->sProcessar && spont->Sock>=0)
            {
                    // Tenta conectar, chama evento
                tamsock=sizeof(SockStr_in);
                localSocket=accept(spont->Sock,
                    (struct sockaddr *)&SockStr_in, &tamsock);
#ifdef __WIN32__
                if (localSocket==INVALID_SOCKET)
                    break;
                int sopcoes;
                int stamanho;
                unsigned long argp=1; // 0=bloquear  1=nao bloquear
                ioctlsocket(localSocket, FIONBIO, &argp);
                sopcoes=2048;
                stamanho=sizeof(sopcoes);
                setsockopt(localSocket, SOL_SOCKET, SO_RCVBUF,
                        (const char*)(void*)&sopcoes, stamanho);
                sopcoes=2048;
                stamanho=sizeof(sopcoes);
                setsockopt(localSocket, SOL_SOCKET, SO_SNDBUF,
                        (const char*)(void*)&sopcoes, stamanho);
#else
                if (localSocket<0)
                    break;
                size_t sopcoes;
                acceptarg3 stamanho;
                fcntl(localSocket, F_SETFL, O_NONBLOCK);
                sopcoes=2048;
                stamanho=sizeof(sopcoes);
                setsockopt(localSocket, SOL_SOCKET, SO_RCVBUF, &sopcoes, stamanho);
                sopcoes=2048;
                stamanho=sizeof(sopcoes);
                setsockopt(localSocket, SOL_SOCKET, SO_SNDBUF, &sopcoes, stamanho);
#endif
                spont->sEvento(localSocket, &SockStr_in);
                if (PontAtual==0)
                    break;
            }
        }
        spont=PontProximo;
    }
}

//------------------------------------------------------------------------------
// Chamado internamente: Prepara TRelogio e TComunica
void ProcessarPrepara(void)
{
    Inic=false;
    memset(TRelogio::Inicio, 0, sizeof(TRelogio*)*TRELOGIO_MAX*3);
    memset(TComunica::Inicio, 0, sizeof(TComunica*)*6);
    memset(TComunica::Fim, 0, sizeof(TComunica*)*6);
#ifdef __WIN32__
        // Inicializa WinSock
        // Nao deve ser chamado aqui porque ja foi chamado em main()
  //  WSADATA info;
  //  bool xinic=(WSAStartup(MAKELONG(1, 1), &info) != SOCKET_ERROR);
#else
        // Ignora sinal PIPE
    signal(SIGPIPE, SIG_IGN);
#endif
}

//------------------------------------------------------------------------------
void ProcessarFim(void)
{
    Term=Processando;
}

//------------------------------------------------------------------------------
void ProcessarIni(void)
{
    struct timeval tselect,tnovo;
    fd_set set_entrada;     // Flags de entrada do select
    fd_set set_entrada2;
    fd_set set_saida;       // Flags de sada do select
    fd_set set_saida2;
    TComunica * cpont;
    TServidor * spont;
    unsigned short atual,tempo; // Para calcular tempo decorrido
    unsigned short espera;
    int cont;              // Contador, para varrer lista de nveis de objetos
    if (Processando)
    {
        Term=false;
        return;
    }
    if (Inic)
        ProcessarPrepara();
    Processando=true;
#ifdef __WIN32__
    atual=timeGetTime()/200;
#else
    gettimeofday(&tselect,0);
    atual = (tselect.tv_sec&0xFFFF)*5 + tselect.tv_usec/200000;
#endif
    FD_ZERO(&set_entrada);
    FD_ZERO(&set_saida);
    while (!Term)
    {
    // Obtm a hora
        time_t tempoatual=0;
        struct tm * tempolocal;
        time(&tempoatual);
        Tempo=tempoatual;
        // localtime() Converte para representaao local de tempo
        tempolocal=localtime(&tempoatual);
        Dia=tempolocal->tm_yday+1;
        Hora=tempolocal->tm_hour;
        Min=tempolocal->tm_min;
    // Data e hora para o registro
        HoraReg=tempolocal->tm_year+300; // Anos, desde 1900
                                // Somando 300 para ser mltiplo de 400
        HoraReg--;
        HoraReg=HoraReg*365+HoraReg/4-HoraReg/100+HoraReg/400;
        HoraReg+=tempolocal->tm_yday; // Dias completos desde o incio do ano
        HoraReg=HoraReg*1440+Hora*60+Min;
        //printf("%d\n", (int)HoraReg); fflush(stdout);
    // Relgio: Calcula o tempo decorrido
        espera=atual;
#ifdef __WIN32__
        atual = timeGetTime()/200;
#else
        gettimeofday(&tselect,0);
        atual = (tselect.tv_sec&0xFFFF)*5 + tselect.tv_usec/200000;
#endif
        espera=atual-espera;
        if (espera>0x800) espera=0;
        if (espera>TESPERA_MAX) espera=TESPERA_MAX;
    // Relgio: Chama eventos
        if (espera>0)
            TempoIni += espera;
        TRelogio::ProcEventos(espera);
    // Verifica conexes pendentes
        TServidor::ProcEventos(&set_entrada);
    // TComunica: Chama eventos
        TComunica::ProcEventos(&set_entrada, 6);
    // TComunica: Envia dados pendentes
        for (cont=0; cont<6; cont++)
        {
            char nivel=TComunica::NivelProc;
            TComunica::NivelProc=cont;
            for (cpont=TComunica::Inicio[cont]; cpont;)
            {
                TComunica::PontProximo[cont]=cpont->Proximo;
                if (cpont->NivelEnvia<5)
                  if (cpont->cSockEnv>=0 && cpont->Leitura!=cpont->Escrita)
                    if (cpont->cEnviaPend())
                    {
                        TComunica::NivelProc = cont;
                        cpont->cEvento(0,1);
                    }
                cpont=TComunica::PontProximo[cont];
            }
            TComunica::NivelProc=nivel;
        }
    // Prepara variveis para select()
        FD_ZERO(&set_entrada);
        FD_ZERO(&set_saida);
        for (cont=0; cont<5; cont++)
        {
            for (cpont=TComunica::Inicio[cont]; cpont; cpont=cpont->Proximo)
            {
                if (cpont->cSockRec>=0)
                    FD_SET(cpont->cSockRec, &set_entrada);
                if (cpont->cSockEnv>=0 && cpont->Leitura!=cpont->Escrita)
                    FD_SET(cpont->cSockEnv, &set_saida);
            }
            if (cont==3)
            {
                for (spont=TServidor::Inicio; spont; spont=spont->Proximo)
                    if (spont->Sock>=0 && spont->sProcessar)
                        FD_SET(spont->Sock, &set_entrada);
                memcpy(&set_entrada2, &set_entrada, sizeof(set_entrada));
                memcpy(&set_saida2, &set_saida, sizeof(set_saida));
            }
        }
        for (cpont=TComunica::Inicio[5]; cpont; cpont=cpont->Proximo)
            if (cpont->cSockRec>=0)
                FD_SET(cpont->cSockRec, &set_entrada);
    // Relgio: Obtm quanto tempo deve esperar
        espera=1;
        while (espera<TESPERA_MAX &&
                TRelogio::Inicio[(TRelogio::TempoAtual1+espera)%(TRELOGIO_MAX*2)]==0)
            espera++;
#ifdef __WIN32__
        tempo = timeGetTime()/200;
#else
        gettimeofday(&tselect,0);
        tempo = (tselect.tv_sec&0xFFFF)*5 + tselect.tv_usec/200000;
#endif
        tempo-=atual;
        if (espera<tempo)
            espera=0;
        else
            espera-=tempo;
    // Espera
        tselect.tv_sec = espera/10;
        tselect.tv_usec = espera%10*100000;
        tnovo.tv_sec=0;
        tnovo.tv_usec=30000;
        select(FD_SETSIZE, &set_entrada2, &set_saida2, 0, &tnovo);
        select(FD_SETSIZE, &set_entrada, &set_saida, 0, &tselect);
 //putchar('#'); fflush(stdout);
    }
    Processando=false;
    Term=false;
}
