#ifndef EVENTOS_H
#define EVENTOS_H

#ifdef __WIN32__
 #include <windows.h>
 #include <winsock.h>
#else
 #include <sys/types.h>  // fd_set no Linux
 #include <sys/time.h>   // fd_set no AIX
 #include <netinet/in.h> // Contm a estrutura sockaddr_in
#endif

#define TRELOGIO_MAX 300    // Tempo mximo em segundos:
                            // TRELOGIO_MAX * TRELOGIO_MAX / 10
#define TESPERA_MAX  12     // Tempo mximo que pode esperar
                            // Deve ser menor que TRELOGIO_MAX
#define COMUNICA_BUF 2048   // Tamanho do buffer de transmisso de TComunica
#define COMUNICA_REC 2048   // Tamanho do buffer para ler mensagem de TComunica
                            // No ocupa espao na memria de TComunica

//------------------------------------------------------------------------------
void Terminar(int retorno); // Deve ser chamado para terminar o programa

//------------------------------------------------------------------------------
// Objeto para processar eventos gerados aps algum tempo
// Os tempos so: 5 = 1 segundo
// Eventos podem ser chamados daqui no mximo 5 horas
class TRelogio
{
public:
    TRelogio();
    virtual ~TRelogio();
    virtual void rEvento();
    bool rPendEvento();         // Retorna verdadeiro se tem evento pendente
    int rTempo(void);           // Informa quanto tempo falta para chamar evento
                                // Retorna 0 se evento no ser chamado
    int rTempo(int NovoTempo);  // Muda o tempo para chamar evento
                                // Entrada: Se <=0, evento no ser chamado
                                // Retorna o mesmo que rTempo(void)
private:
    static void TRelogio::ProcEventos(int TempoDecorrido);
    TRelogio * Anterior;        // Objetos anterior e prximo, na lista ligada
    TRelogio * Proximo;
    short Indice;               // Se <0, objeto no est esperando um evento
                                // Se ==TRELOGIO_MAX, o evento ser chamado
                                // Se >=0, Relogio[Tempo] aponta para
                                // o primeiro objeto da lista ligada
    short Indice2;              // Se >=0,  para quanto vai Indice1 quando
                                // o tempo de Indice1 se esgotar
    static TRelogio * Inicio[TRELOGIO_MAX * 3];
    static int TempoAtual1;     // Ponteiro Relogio correspondente ao tempo atual
    static int TempoAtual2;

    friend void ProcessarPrepara();
    friend void ProcessarIni(void);
};

//------------------------------------------------------------------------------
// Objeto para manipular conexes
// No cria conexes, mas fecha "socket" se objeto apagado ou conexo terminada
// cEvento() - evento para receber mensagens:
//   Se buffer==0 e tamanho==0: conexo terminada pelo outro lado
//                              todos os "socket"s foram fechados
//   Se buffer==0 e tamanho==1: todos os dados no buffer de sada foram enviados
//   Se buffer!=0: recebeu "tamanho" bytes, esto em "buffer"
// cEnvia() - funo para enviar mensagens:
//   Entrada:  endereo do buffer, tamanho do buffer
//             Se o tamanho for omitido ou for <0, o buffer  uma string
//             (o byte =0 da string no  enviado)
//   Retorna:  verdadeiro: conseguiu enviar
//             falso: no conseguiu enviar (nada foi enviado), motidos:
//                    - o buffer de transmisso est cheio; ou
//                    - foi solicitado mais de COMUNICA_BUF bytes; ou
//                    - no pode enviar devido ao nvel (ver nveis, abaixo)
// Nveis:
//   Os nveis dizem respeito ao funcionamento de cEnvia() e cEvento()
//   H dois nveis: para cEnvia() e para cEvento()
//   Quando o objeto  criado, o nvel de ambos  4
// Nveis de cEnvia():
// 5     cEnvia() coloca a mensagem no buffer de envio, ao invs de enviar
//       Quando o buffer ficar cheio, o nvel de cEnvia() muda para 4,
//       e cEnvia() passa a enviar mensagens normalmente
// 4     cEnvia() tenta enviar mensagem e retorna imediatamente avisando
//       se conseguiu
// 3-0   cEnvia() no envia mensagem (retorna falso) se for chamado
//       pela funo cEvento() e nvel de cEnvia() > nvel de cEvento()
//       Nos demais casos, cEnvia() tenta enviar a mensagem e no retorna at
//       conseguir enviar. Enquanto isso, eventos cEvento() de menor nvel
//       podero ocorrer.
class TComunica
{
public:
    TComunica();
    virtual ~TComunica();
    virtual void cEvento(const char * buffer, int tamanho);
    short cSockEnv;             // "Socket" para enviar e para receber,
    short cSockRec;             // -1 significa que no est sendo usado
    void cFechaSock(bool enviar=true);// Fecha "socket"s cSockEnv e cSockRec, NO GERA EVENTO
                                // Se enviar=true, envia dados pendentes no buffer de sada
    bool cEnvia(const void * buffer, int tamanho=-1);
                                // Envia mensagem, retorna verdadeiro
                                // se conseguiu enviar
    bool cEnviaPend(void);      // Envia dados pendentes no buffer de sada, NO GERA EVENTO
                                // Chamado internamente (o programa no precisa chamar)
                                // Retorna verdadeiro se buffer vazio (no gera evento)
    char cNivelEnvia(void);     // Retorna o nvel
    char cNivelEvento(void);
    void cMudaNivel(char envia, char evento);// Muda os nveis (ver acima)
    void cVoltaMens(void);      // Volta mensagens de 3 segundos atrs
                                // Utilizado em conexes com o bate-papo via telnet
    void cLimpaMens(void);      // Cancela as ltimas mensagens, de modo que cVoltaMens()
                                // no possa recuper-las
    char * cBuffer(void);       // Fornece o endereo da matriz Buffer
                                // Essa matriz pode ser usada se TComunica
                                // no for usado para transmitir dados

private:
    static void ProcEventos(fd_set * set_entrada, char TotalNiveis); // Gera eventos
    bool  Fechou;               // Verdadeiro se deve avisar que conexo terminou
    unsigned char NivelEnvia;   // Nvel de cEnvia()  (ver acima)
    unsigned char NivelEvento;  // Nvel de cEvento() (ver acima)
    char  Buffer[COMUNICA_BUF]; // Buffer com dados que sero enviados
    short Leitura,Escrita;      // Ponteiros de leitura e de escrita no buffer
    short Esc0Seg, Esc1Seg, Esc2Seg, Esc3Seg;// Ponteiro de escrita aps decorrido
                                // de 0 a 3 segundos, -1 significa ponteiro no usado
    long  EscTempo;             // Tempo de Esc0Seg

    TComunica * Anterior;       // Objetos anterior e prximo, na lista ligada
    TComunica * Proximo;

                                // As prximas variveis referem-se a NivelEvento
    static TComunica * Inicio[6];// Apontam para primeiro objeto de nveis 0 a 5
    static TComunica * Fim[6];   // Apontam para o ltimo objeto
    static TComunica * PontAtual[6]; // Apontam para o objeto da lista
                                // (nveis 0 a 5) que est sendo processado
                                // Permitem saber quando um objeto  apagado
    static TComunica * PontProximo[6]; // Apontam para o prximo objeto da lista
                                // (nveis 0 a 5) que ser processado
    static char NivelProc;      // Nvel em que est, ao chamar cEvento()
                                // cEnvia() retorna falso se NivelEnvia>NivelProc
                                // e NivelEnvia<4. Por isso, NivelProc comea com 6

    friend void ProcessarPrepara();
    friend void ProcessarIni(void);
};

//------------------------------------------------------------------------------
// Objeto para receber conexes
class TServidor
{
public:
    TServidor();
    virtual ~TServidor();
    char * sAtivar(unsigned short * Porta);
                                // Faz objeto esperar por conexo em uma porta
                                // Fornecer a porta, 0 significa que ser
                                //     escolhida uma porta disponvel
                                // Retorna end. da mensagem de erro,
                                //     retorna 0 se conseguiu abrir a porta
    void sDesativar(void);      // Fecha a porta
    virtual void sEvento(short EvSock, struct sockaddr_in * localSockStr);
                                // Chamado sempre que chegar uma conexo
                                // EvSock contm o descritor do "socket"
    static void sProcTodos(bool Processar);// Ativa ou desativa a varivel processar
                                // de todos os objetos do processo
    bool sProcessar;            // Se falso, no recebe conexes

private:
    static void TServidor::ProcEventos(fd_set * set_entrada);
    short Sock;                 // "socket" esperando conexo, <0 se no houver
    TServidor * Anterior;       // Aponta para objeto anterior
    TServidor * Proximo;        // Aponta para prximo objeto
    static TServidor * Inicio;  // Aponta para primeiro objeto da lista
    static TServidor * PontAtual;// Objeto que est sendo processado
    static TServidor * PontProximo;// Prximo objeto que ser processado

    friend void ProcessarIni(void);
};

//------------------------------------------------------------------------------
// Processa objetos TComunica e TRelogio, s retorna quando
// ProcessaFim() for chamado.
// Se chamado de um evento de TComunica ou TRelogio, cancela a ao de
// ProcessarFim(), e retorna imediatamente.
void ProcessarIni(void);

// Termina de processar, s tem efeito se chamado de um evento
// de TComunica ou TRelogio
void ProcessarFim(void);

// Chamado internamente por evtempo.cpp
void ProcessarPrepara(void);

// Tempo atual (em segundos) - disponvel para os eventos
extern long Tempo;

// Para contar tempo decorrido (varinc e vardec)
extern unsigned long TempoIni;

// Hora atual
extern short Dia,Hora,Min;

// Nmeros que definem o dia e hora/minuto
// Usados no registro
extern unsigned long HoraReg;

#endif
